package ZSCII;
{
  $ZSCII::VERSION = '0.001';
}
use strict;
use warnings;
# ABSTRACT: libraries for dealing with Z-Machine text


1;

__END__

=pod

=head1 NAME

ZSCII - libraries for dealing with Z-Machine text

=head1 VERSION

version 0.001

=head1 OVERVIEW

The ZSCII distribution contains libraries for encoding and decoding text from
Z-Machine programs.

=head1 AUTHOR

Ricardo SIGNES <rjbs@cpan.org>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2013 by Ricardo SIGNES.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
